/*
* decompress.cpp - Decompression and compression algorithms for Faxanadu level data.
*
* Copyright (c) 2005 Sebastian Porst (webmaster@the-interweb.com)
* All rights reserved.
*
* This software is licensed under the zlib/libpng License.
* For more details see http://www.opensource.org/licenses/zlib-license.php
*/

#include <iostream>
#include <iterator>
#include <cassert>
#include <vector>
#include <iomanip>
#include <fstream>

#include "neshelp.h"

struct Screen
{
	unsigned char tiles[13][16];
	// The next line is only used for testing compression.
	StuffBits<2000> cdata;
};

struct Level
{
	std::vector<Screen> screens;
};

std::ostream& operator<<(std::ostream& os, const Screen& screen)
{
	for (unsigned int i=0;i<13;i++)
	{
		for (unsigned int j=0;j<16;j++)
		{
			os << std::hex << std::setfill('0') << std::setw(2) << (unsigned int)screen.tiles[i][j] << " ";
		}
		
		os << "\n";
	}
	
	return os;
}

std::ostream& operator<<(std::ostream& os, const Level& l)
{
	std::copy(l.screens.begin(), l.screens.end(), std::ostream_iterator<Screen>(os, "\n"));
	return os;
}

unsigned char getTile(unsigned int& bit, unsigned char* data, Screen& screen, unsigned int x, unsigned int y)
{
	unsigned int hbits = getBits(data,  bit / 8, bit % 8, 2);
	
	// The next line is only used for testing compression.
	screen.cdata.addBits(hbits, 2);
	
	unsigned int nextByte = (bit + 2) / 8;
	unsigned int nextBit = (bit + 2) % 8;
	
	bit += 2;
	
	switch(hbits)
	{
		case 0:
			if (x == 0) {assert(y); return screen.tiles[y-1][15];}
			else return screen.tiles[y][x-1];
		case 1:
			return screen.tiles[y-1][x];
		case 2:
			if (x == 0) {return screen.tiles[y-2][15];}
			return screen.tiles[y-1][x-1];
		case 3:
			bit += 8;
			// The next line is only used for testing compression.
			screen.cdata.addBits(getBits(data, nextByte, nextBit, 8), 8);
			return getBits(data, nextByte, nextBit, 8);
	}
	
	assert(0);
}

unsigned int decompressScreen(unsigned char* data, Screen& screen)
{
	unsigned int bit = 0;
	for (unsigned int i=0;i<13;i++)
	{
		for (unsigned int j=0;j<16;j++)
		{
			screen.tiles[i][j] = getTile(bit, data, screen, j, i);
		}
	}

	return bit;
}

void decompressChunk(unsigned char* data, unsigned int baseoffset, std::vector<Screen>& screens)
{
	unsigned short offset_first_chunk = *(unsigned short*)data;
	
	unsigned int nrtiles = (offset_first_chunk - baseoffset) / 2;

	for (unsigned int i=0;i<nrtiles;i++)
	{
		Screen screen;

		unsigned short screenpointer = (*(unsigned short*)(data + i * 2)) - baseoffset;
		decompressScreen(data + screenpointer, screen);
		
		screens.push_back(screen);
	}
}

void decompressLevel(unsigned char* data, std::vector<Level>& levels)
{
	unsigned short nrlevels = (*(unsigned short*)data) / 2;
	
	for (unsigned int i=0;i<nrlevels;i++)
	{
		Level l;
		unsigned short lvlpointer = (*(unsigned short*)(data + i * 2));
		decompressChunk(data + lvlpointer, lvlpointer, l.screens);

		levels.push_back(l);
	}
}

template<unsigned int buffersize>
void compressTile(const Screen& screen, unsigned int x, unsigned int y, StuffBits<buffersize>& buffer)
{
	unsigned char tile = screen.tiles[y][x];
	
	if (x == 0 && y > 0 && screen.tiles[y-1][15] == tile) buffer.addBits(0,2);
	else if (x > 0 && screen.tiles[y][x-1] == tile) buffer.addBits(0, 2);
	else if (y > 0 && screen.tiles[y-1][x] == tile) buffer.addBits(1, 2);
	else if (x > 0 && y > 0 && screen.tiles[y-1][x-1] == tile) buffer.addBits(2, 2);
	else if (x == 0 && y > 1 && screen.tiles[y-2][15] == tile) buffer.addBits(2, 2);
	else
	{
		buffer.addBits(3, 2);
		buffer.addBits(tile, 8);
	}
}

// 16x13x8
template<unsigned int buffersize>
void compressScreen(const Screen& screen, StuffBits<buffersize>& buffer)
{
	for (unsigned int y=0;y<13;y++)
	{
		for (unsigned int x=0;x<16;x++)
		{
			compressTile(screen, x, y, buffer);
		}
	}
}

int main()
{
	std::string romname = "Faxanadu (U).nes";
	
	std::vector<Level> l;
	
	unsigned int chunkOffsets[] = {0x0010, 0x4010, 0x8010};
	
	unsigned char* lvlData = new unsigned char[0x4000];
	std::ifstream file(romname.c_str(), std::ios_base::binary);
	
	for (unsigned int i=0;i<sizeof(chunkOffsets)/sizeof(*chunkOffsets);i++)
	{
		file.seekg(chunkOffsets[i]);
		file.read(reinterpret_cast<char*>(lvlData), 0x4000);
		decompressLevel(lvlData, l);
	}
	
	delete[] lvlData;
	
	for (unsigned int i=0;i<l.size();i++)
	{
		for (unsigned int j=0;j<l[i].screens.size();j++)
		{
			StuffBits<260> cdata;
			compressScreen(l[i].screens[j], cdata);
			unsigned int size = cdata.size();
			unsigned int size2 = l[i].screens[j].cdata.size();
			assert(size % 8 == 0);
			
			for (unsigned int k=0;k<size / 8;k++)
			{
				assert(cdata.getBuffer()[k] == l[i].screens[j].cdata.getBuffer()[k]);
			}
			assert(size == size2);
		}
	}
	
	std::copy(l.begin(), l.end(), std::ostream_iterator<Level>(std::cout, "\n"));
}
